@@ -21,6 +21,8 @@
 #include "layout/htmlboxtext.h"
 #include "layout/htmlboxinline.h"
 #include "htmlviewaccessible.h"
+#include "htmllinkaccessible.h"
+#include <util/gail-util.h>
 
 static void           html_box_text_accessible_class_init          (HtmlBoxTextAccessibleClass *klass);
 static void           html_box_text_accessible_finalize            (GObject             *object);
@@ -101,9 +103,16 @@ static AtkHyperlink*  html_box_text_accessible_get_link        (AtkHypertext
 static gint           html_box_text_accessible_get_n_links     (AtkHypertext      *hypertext);
 static gint           html_box_text_accessible_get_link_index  (AtkHypertext      *hypertext,
                                                                 gint              char_index);
+static gboolean       has_link                                 (AtkHypertext      *hypertext);
 
 static gpointer parent_class = NULL;
 
+struct _HtmlBoxTextAccessiblePrivate
+{
+	GailTextUtil *textutil;
+	gpointer links;
+};
+
 GType
 html_box_text_accessible_get_type (void)
 {
@@ -171,14 +180,16 @@ html_box_text_accessible_real_initialize (AtkObject *object,
 	ATK_OBJECT_CLASS (parent_class)->initialize (object, data);
 
 	text = HTML_BOX_TEXT_ACCESSIBLE (object);
+        text->priv = g_new0 (HtmlBoxTextAccessiblePrivate, 1);
 	box_text = HTML_BOX_TEXT (data);
 	contents  = html_box_text_get_text (box_text, &len);
 	text_buffer = gtk_text_buffer_new (NULL);
 	if (contents)
 		gtk_text_buffer_set_text (text_buffer, contents, len);
-	text->textutil = gail_text_util_new ();
-	gail_text_util_buffer_setup (text->textutil, text_buffer);
+	text->priv->textutil = gail_text_util_new ();
+	gail_text_util_buffer_setup (text->priv->textutil, text_buffer);
 	g_object_unref (text_buffer);
+
 }
 
 static void
@@ -199,7 +210,11 @@ html_box_text_accessible_finalize (GObject *object)
 {
 	HtmlBoxTextAccessible *text = HTML_BOX_TEXT_ACCESSIBLE (object);
 
-	g_object_unref (text->textutil);
+	if (text->priv->links)
+		g_object_unref (text->priv->links);
+
+	g_object_unref (text->priv->textutil);
+	g_free (text->priv);
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -238,9 +253,9 @@ html_box_text_accessible_get_text (AtkText *text,
 
 	g_return_val_if_fail (HTML_IS_BOX_TEXT_ACCESSIBLE (text), NULL);
 	text_accessible = HTML_BOX_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (text_accessible->textutil, NULL);
+	g_return_val_if_fail (text_accessible->priv->textutil, NULL);
 
-	buffer = text_accessible->textutil->buffer;
+	buffer = text_accessible->priv->textutil->buffer;
 	gtk_text_buffer_get_iter_at_offset (buffer, &start, start_offset);
 	gtk_text_buffer_get_iter_at_offset (buffer, &end, end_offset);
 
@@ -295,7 +310,7 @@ html_box_text_accessible_get_character_at_offset (AtkText *text,
 
 	g_return_val_if_fail (HTML_IS_BOX_TEXT_ACCESSIBLE (text), 0);
 	text_accessible = HTML_BOX_TEXT_ACCESSIBLE (text);
-	buffer = text_accessible->textutil->buffer;
+	buffer = text_accessible->priv->textutil->buffer;
 	gtk_text_buffer_get_start_iter (buffer, &start);
 	gtk_text_buffer_get_end_iter (buffer, &end);
 	string = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);
@@ -313,8 +328,8 @@ html_box_text_accessible_get_character_count (AtkText *text)
 
 	g_return_val_if_fail (HTML_IS_BOX_TEXT_ACCESSIBLE (text), 0);
 	text_accessible = HTML_BOX_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (text_accessible->textutil, 0);
-	buffer = text_accessible->textutil->buffer;
+	g_return_val_if_fail (text_accessible->priv->textutil, 0);
+	buffer = text_accessible->priv->textutil->buffer;
 	return gtk_text_buffer_get_char_count (buffer);
 }
 
@@ -328,8 +343,8 @@ html_box_text_accessible_get_caret_offset (AtkText *text)
 
 	g_return_val_if_fail (HTML_IS_BOX_TEXT_ACCESSIBLE (text), 0);
 	text_accessible = HTML_BOX_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (text_accessible->textutil, 0);
-	buffer = text_accessible->textutil->buffer;
+	g_return_val_if_fail (text_accessible->priv->textutil, 0);
+	buffer = text_accessible->priv->textutil->buffer;
 	cursor_mark = gtk_text_buffer_get_insert (buffer);
 	gtk_text_buffer_get_iter_at_mark (buffer, &cursor_itr, cursor_mark);
 	return gtk_text_iter_get_offset (&cursor_itr);
@@ -345,8 +360,8 @@ html_box_text_accessible_set_caret_offset (AtkText *text,
 
 	g_return_val_if_fail (HTML_IS_BOX_TEXT_ACCESSIBLE (text), FALSE);
 	text_accessible = HTML_BOX_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (text_accessible->textutil, FALSE);
-	buffer = text_accessible->textutil->buffer;
+	g_return_val_if_fail (text_accessible->priv->textutil, FALSE);
+	buffer = text_accessible->priv->textutil->buffer;
 	gtk_text_buffer_get_iter_at_offset (buffer,  &pos_itr, offset);
 	gtk_text_buffer_move_mark_by_name (buffer, "insert", &pos_itr);
 	return TRUE;
@@ -379,8 +394,12 @@ html_box_text_accessible_get_offset_at_point (AtkText      *text,
 	box_text = HTML_BOX_TEXT (g_obj);
 	box = HTML_BOX (g_obj);
 
-	if (box->prev == NULL && HTML_IS_BOX_INLINE (box->parent))
-		x -= html_box_left_border_width (box->parent);
+	if (box->prev == NULL) {
+		while (HTML_IS_BOX_INLINE (box->parent)) {
+			x -= html_box_left_border_width (box->parent);
+			box = box->parent;
+		}
+	}
 	return html_box_text_get_index (box_text, x - real_x);
 }
 
@@ -413,8 +432,12 @@ html_box_text_accessible_get_character_extents (AtkText      *text,
 
 	html_box_text_get_character_extents (box_text, offset, &rect);
         *x = real_x + rect.x;
-	if (box->prev == NULL && HTML_IS_BOX_INLINE (box->parent))
-		*x += html_box_left_border_width (box->parent);
+	if (box->prev == NULL) {
+		while (HTML_IS_BOX_INLINE (box->parent)) {
+			*x += html_box_left_border_width (box->parent);
+			box = box->parent;
+		}
+	}
 	
         *y = real_y;
         *width = rect.width;
@@ -778,7 +801,7 @@ get_text_near_offset (AtkText          *text,
                       gint             *start_offset,
                       gint             *end_offset)
 {
-	return gail_text_util_get_text (HTML_BOX_TEXT_ACCESSIBLE (text)->textutil, NULL,
+	return gail_text_util_get_text (HTML_BOX_TEXT_ACCESSIBLE (text)->priv->textutil, NULL,
 					function, boundary_type, offset, 
 					start_offset, end_offset);
 }
@@ -797,18 +820,70 @@ static AtkHyperlink*
  html_box_text_accessible_get_link (AtkHypertext *hypertext,
                                     gint         link_index)
 {
-	return NULL;
+	HtmlBoxTextAccessible *text;
+
+	text = HTML_BOX_TEXT_ACCESSIBLE (hypertext);
+	if (has_link (hypertext)) {
+		if (!text->priv->links) {
+			text->priv->links =  html_link_accessible_new (ATK_OBJECT (hypertext));
+		}
+		return text->priv->links;
+	}
+	else
+		return NULL;
 }
 
 static gint
 html_box_text_accessible_get_n_links (AtkHypertext *hypertext)
 {
-	return 0;
+	if (has_link (hypertext))
+		return 1;
+	else
+		return 0;		
 }
 
 static gint
 html_box_text_accessible_get_link_index (AtkHypertext *hypertext,
                                          gint         char_index)
 {
-	return -1;
+	gint index = -1;
+
+	if (has_link (hypertext)) {
+		gint char_count;
+
+		char_count = html_box_text_accessible_get_character_count (ATK_TEXT (hypertext));
+		if (char_count > 0 && char_index >= 0 && char_index < char_count)
+			index = 0;
+	}
+	return index;
 }
+
+static gboolean
+has_link (AtkHypertext *hypertext)
+{
+	AtkGObjectAccessible *atk_gobj;
+	GObject *g_obj;
+	HtmlBox *box;
+	gboolean has_link = FALSE;
+
+	atk_gobj = ATK_GOBJECT_ACCESSIBLE (hypertext);
+	g_obj = atk_gobject_accessible_get_object (atk_gobj);
+	if (g_obj == NULL)
+		return has_link;
+
+	box = HTML_BOX (g_obj);
+	if (HTML_IS_BOX_INLINE (box->parent)) {
+		DomNode *node;
+
+		node = box->parent->dom_node;
+
+		if (node->xmlnode->name) {
+			if (strcasecmp (node->xmlnode->name, "a") == 0 &&
+			   (xmlHasProp (node->xmlnode, "href") != NULL)) {
+				has_link = TRUE;
+			}
+		}
+	}
+	return has_link;
+}
+
